<?php

/**
 * @file
 * Home of the Netforum Feeds and related classes.
 */

// netforum will only return a maximum of 500 results at a time when using certain
// functions. this means we'll have to iterate in batches. this value 
define('NETFORUM_MAX_RESULTS', 50);

/**
 * Result of FeedsHTTPFetcher::fetch().
 */
class NetforumConnectionResult extends FeedsFetcherResult {
  /**
   * Constructor.
   */
  public function __construct($object_type, $arguments) {
    $this->object_type=$object_type;
    $this->arguments=$arguments;
    
    parent::__construct('');
  }

  /**
   * Overrides FeedsFetcherResult::getRaw();
   */
  public function getRaw() {
    // prevent this from running successive times
    if(!isset($this->response)){
      $this->response = netforum_xweb_request($this->object_type, $this->arguments);
      
      // reduce the returned response to max_results. this is usefull for testing date iterations
      // wont be practical in realworld scenarios
      $test=(array)$this->response;
      if(count($test['Result']) > NETFORUM_MAX_RESULTS){
        $test['Result']=array_slice($test['Result'], 0, NETFORUM_MAX_RESULTS);
        $this->response=$test;
      }
    }
    
    return $this->response;
  }
}

/**
 * Fetches data from Netforum.
 */
class NetForumConnection extends FeedsFetcher {

  /**
   * Implements FeedsFetcher::fetch().
   */
  public function fetch(FeedsSource $source) {
    $state=$source->state(FEEDS_FETCH);
  
    // get the configuration for this xweb request
    $config=$this->getFetchConfig();

    // if the feed is configured to batch-iterate set it up here
    if(count($config['iterate'])){
      foreach($config['iterate'] as $iterate_key=>$iterate_type){
        if(!isset($state->iterate)){
          // iteration can happen across several grouped fields
          // we need to prepare for that, but for now there's just one field
          $state->iterate=array();
          
          // set the default iteration values
          $state->iterate[$iterate_key]=$config['attributes'][$iterate_key];
          
          // modify the iterate config element's start variable to match batch start date.
          // that value changes incrementally as the data comes in, and is modified at the
          // end of the function
        }
        
        // iteration is going to happen, populate iteratable values into config attributes
        // before the query is run
        $config['attributes'][$iterate_key]=$state->iterate[$iterate_key];
      }
    }
    
    // get the data needed here
    $fetcher=new NetforumConnectionResult($this->config['retrieve_object'], $config['attributes']);
    
    $data=(array)$fetcher->getRaw();
    //only iterate if we have iterators. collect the number of them for processing
    if($i=count($config['iterate'])){
      // check in here for the iterator to increment. there are a couple of basic
      // rules. we'll reverse the order of iterator, and if it can determine that
      // it should go to the next increment it will break the loop.
      //$i keeps track of the depth of the iterator chain (probably only ever 2 deep)
      foreach(array_reverse($config['iterate']) as $iterate_key=>$iterate_type){
        $i--;
        if($iterate_type=='date_field'){
          //if it's a date field, then the increment is by record count
          //TODO: make it clear that this is different from a date range field
          if(count($data['Result']) >= NETFORUM_MAX_RESULTS){
            // the results that came back were limited which means we have to adjust
            // this date field and move forward
            break;
          }
          //as long as $i > 0 this field can be reset for the next value to iterate with
          //if this doesn't happen, an infinite loop will result.
          elseif($i){
            // if we get here, iteration doesn't need to take place, so the field
            // should be reset back to default
            $state->iterate[$iterate_key]=$config['attributes'][$iterate_key];
          }
          
        }
        // the only other option is alphabetical right now
        else{
          // don't reset if we are on the last key
          if($state->iterate[$iterate_key] < 'z'){
            // we haven't reached z yet.
            break;
          }
          //as long as $i > 0 this field can be reset for the next value to iterate with
          //if this doesn't happen, an infinite loop will result.
          elseif($i){
            $state->iterate[$iterate_key]=$config['attributes'][$iterate_key];
          }
        }        
      }
      
      // handle alphabetical iteration first
      if($iterate_type=='alphabetical'){
        if($state->iterate[$iterate_key] < 'z'){
          //increment to the next letter for searching
          $state->iterate[$iterate_key]=chr(ord(strtolower($state->iterate[$iterate_key]))+1);
          $state->progress(ord('z')-ord('a'), ord($state->iterate[$iterate_key])-ord('a'));
          
        }
      }
      elseif($iterate_type=='date_field'){
        //get the mappings
        $processor_config=$source->importer->processor->getConfig();
        
        foreach($processor_config['mappings'] as $index=>$mapelement){
          if($mapelement['target'] == 'iteratorField'){
            $iterator_field=$mapelement['source'];
            break;
          }
        }
        // if the iterator field has been defined
        if($iterator_field){
          //TODO: confirm that the last record is the newest.
          $last_row=count($data['Result'])-1;
          //correct the starting date so proper(ish) batch completion progress can be reported
          if(!isset($state->batch_start_date)){
            
            // note: using unix timestamps
            $state->batch_start_date=strtotime($data['Result'][0]->{$iterator_field});
          }
          
          // if netforum returned the maximum records allowed for a date query
          // another batch iteration needs to take place.
          if(count($data['Result']) >= NETFORUM_MAX_RESULTS){
            //set the next query-by date to the date of the youngest record in the set
//            $last_row_data=(array)$data['Result'][$last_row];
            $state->iterate[$iterate_key]=(string)$data['Result'][$last_row]->{$iterator_field};
            
            // progress total = now() - earliest_record_date
            // elapsed = next_starting point - earliest_record_date
            $now=strtotime('now');
            $state->progress($now - $state->batch_start_date, strtotime($state->iterate[$iterate_key]) - $state->batch_start_date);
//            $state->progress(1,1);
          }
          // if the record count is w/in the proper range, progress is complete
          else{
            //we're done
            $state->progress(1,1);
          }
          
        }
        
      }
      else{
        // another iterator of some sort
      }
    }
    dpm($state->iterate, 'state iteration values');
    return $fetcher;
  }
  
  
  
  /**
   * Override parent::configDefaults().
   */
  public function configDefaults() {
    $defaults=array(
      'retrieve_object' => '',
    );
    
    foreach(array_keys(_netforum_xweb_soap_functions()) as $function){
      $defaults[$function]=array(
        'iterate'=>array(),
        'attributes'=>array(),
      );
      foreach( netforum_xweb_function_struct_parameters($function) as $param){
        if(!is_array($param)){
          $defaults[$function]['attributes'][$param]=array(
              'type'=>"none",
              'weight'=>0,
              'default_value'=>"",
          );
        }
      }
    }
    
    return $defaults;
  }
  
  /**
   * Override parent::configForm().
   */
  public function configForm(&$form_state) {
    $form = array('#tree'=>true);
    $form['retrieve_object'] = array(
        '#type' => 'select',
        '#title' => t('Object Type to Retrieve'),
        '#description' => t('What kind of Object is this'),
        '#default_value' => $this->config['retrieve_object'],
        '#options'=> array(
        ),
    );
    // Get the soap function list and create fields for each function's params    
    $function_list = _netforum_xweb_soap_functions();
    $i=0;
    foreach (array_keys($function_list) as $function_name) {
      $i++;
//      if($i>=4) break;
      // register the function with the options selection.
      $form['retrieve_object']['#options'][$function_name]=$function_name;
      // iterate through the params list and define any fields that haven't already
      // been defined.
      $form[$function_name]=array(
        '#type'=>'fieldset',
        '#title'=>t('Params for %function', array('%function'=>$function_name)),
        '#dependency'=>array(
          'edit-retrieve-object'=>array($function_name),
        ),
      );
      $params = netforum_xweb_function_struct_parameters($function_name);
//      dpm($params, $function_name);
      $iterate_options=array();
      
      //dont do anything if there aren't any parameters to show
      if(count($params)){
        $form[$function_name]['attributes']=array(
            '#theme'=>'netforum_iteration_table',
            '#function_name'=>$function_name,
        );
        foreach($params as $key=>$option){
          if(!is_array($option)){
            //the param types are in the name of the function
            // this should switch between those types to provide better field entry
            // do this later.
            if(!is_array($this->config[$function_name]['attributes'])){
              $this->config[$function_name]['attributes']=array();
            }
            
            $form[$function_name]['attributes'][$option]=array();
            
            $form[$function_name]['attributes'][$option]['name']=array(
              '#type'=>'item',
              '#title'=>'Field Name',
              '#markup'=>$option,
            );
            $form[$function_name]['attributes'][$option]['default_value']=array(
              '#type'=>'textfield',
              '#title'=>"Default value",
              '#default_value'=>$this->config[$function_name]['attributes'][$option]['default_value'],
              '#description'=>""
            );
            $form[$function_name]['attributes'][$option]['type']=array(
              '#type'=>'select',
              '#title'=>"Iteration method",
              '#options'=>array(
                  'none'=>"Do not iterate with this field",
                  'date_field'=>"Use date field from result set",
                  'alphabetical'=>'Use alphabetical pulling'
              ),
              '#default_value'=>$this->config[$function_name]['attributes'][$option]['type'],
            );
            $form[$function_name]['attributes'][$option]['weight']=array(
              '#type'=>'textfield',
              '#title'=>"Weight",
              '#attributes'=>array('class'=>array()),
              '#default_value'=>$this->config[$function_name]['attributes'][$option]['weight'],
            );
            
          }
        }
      }
    }
    return $form;
  }

  // get the configuration and convert it to the set of directives the fecthing
  // algorithm expects.
  
  public function getFetchConfig(){
    $config=$this->config[$this->config['retrieve_object']];
    
    $fconfig=array();
    //TODO: we should sort by weight first
    foreach($config['attributes'] as $key=>$attribs){
      if($attribs['type'] <> 'none'){
        $fconfig['iterate'][$key]=$attribs['type'];
      }
      $fconfig['attributes'][$key]=$attribs['default_value'];
    }
    
    return $fconfig;
  }
}


class NetforumDatefieldIterator extends NetforumIterator{
  function __construct(&$feed, $value){
    $this->feed=&$feed;
    $this->value=$value;
  }
  function next(&$results){
//    $this->value=ord(asc(lower($this->value))+1);

//    $this->value=date_add()
  }
  function is_complete(){
    if($this->value <= now()) {
      return false;
    }
    return true;
  }
}
class NetforumAlphabeticalIterator extends NetforumIterator{
  function __construct(&$feed, $value){
    $this->feed=&$feed;
    $this->value=$value;
  }
  function next(&$results){
    $this->value=ord(asc(lower($this->value))+1);
  }
  function is_complete(){
    if($this->value <= 'z') {
      return false;
    }
    return true;
  }
}

abstract class NetforumIterator{
  private $feed;
  private $value;
  
  function next(&$results){}
  function is_complete(){}
}
